//-----------------------------------------------------------------------------
// File: Circuit.hh
//
// Purpose: Declaration of the Circuit related structures
//
// History: 03/10/98 - LGS - created.
//
// Copyright (c) 1999 ALGOS Group / INESC
//
//-----------------------------------------------------------------------------

#ifndef __TAU_CIRCUIT__
#define __TAU_CIRCUIT__

#include <iostream.h>
#include <assert.h>

#include "utils/defs.hh"
#include "utils/memory.hh"
#include "utils/darray.hh"
#include "utils/list.hh"
#include "utils/symbTable.hh"

#include "fcgrp_CDB.hh"

#define strsav(s) strcpy((char*)new char[strlen(s)+1],s)
#define LARGE_BUFSIZ 65536

// node types
#define INPUT	0x00    // controllable nodes
#define AND     0x01
#define NAND    0x02
#define OR      0x03
#define NOR     0x04
#define XOR     0x05	// uncontrollable nodes
#define XNOR    0x06
#define BUFF    0x07
#define NOT     0x08
#define ITE	0x09
#define OUTPUT	0x0A
#define BAD     0x0B
#define VIRTUAL 0x0C

// node properties
#define CONTROLLABLE(n) (n->type < XOR)
#define UNCONTROLLABLE(n) (n->type >= XOR)

// check if node n assumes a controlling value
#define CONTROLLED(n) ( n->value() == n->ctrlvalue && CONTROLLABLE(n) )

// check if node m controlls node n
#define CONTROLLING(m,n)    ( ( CONTROLLABLE(n) &&  \
                                m->value() == n->ctrlfanin ) || \
                              ( UNCONTROLLABLE(n) && \
                                m->value() != UNKNOWN ))
#define NONCONTROLLING(m,n) ( CONTROLLABLE(n) &&  \
                              m->value() == 1 - n->ctrlfanin )

// check if node n is assigned
#define ASSIGNED(n)   ( n->value() != -2 )
#define UNASSIGNED(n) ( n->value() == -2 )

// check if node n is justified
#define JUST(n)   ( (*_Jf_data)[n->ID] == NULL && ASSIGNED(n) )
#define UNJUST(n) ( (*_Jf_data)[n->ID] != NULL || UNASSIGNED(n) )

class Node;

typedef Node *NodePtr;
typedef DArray<Node *> DNodeArray;
typedef List<Node *> NodeList;
typedef ListItem<Node *> *NodeListPtr;

class Node {
public:

  // constructor/destructor
  Node (int ID);
  virtual ~Node ();

  // fanin/fanout manipulation
  void add_fanin (Node *fanin) { fanins[fanins.incr_size(1)] = fanin; }
  void add_fanout (Node *fanout) { fanouts[fanouts.incr_size(1)] = fanout; }
  void del_fanin (Node *fanout);
  void del_fanout (Node *fanout);

  int value() { return variable->value(); }
  void unconnect();
  
  // justification counters manipulation 
  void update_fanouts (int value, int count);
  int  is_justified ();

  int ID;
  int type;
  VariablePtr variable;
  char *name;
  DNodeArray fanins;
  DNodeArray fanouts;

  // topological information
  int depth;			// topological depth
  int rdepth;			// reverse topological depth

  // functional information
  int invert;
  int ctrlvalue;
  int ctrlfanin;

  // justification information
  int jc[2];			// justification counters

private:
};

class Circuit {
public:
  Circuit (ClauseDatabase &clDB);
  virtual ~Circuit();
 
  // Node manipulation 
  Node *add_node (char *name);
  void remove_node (int varID) { nodes[varID] = NULL; }
  void remove_node (Node *node) { nodes[node->ID] = NULL; }
  void add_gate (int type, NodePtr output, NodePtr *inputs, int Ninputs); 
  void add_gate (int type, NodePtr output, NodePtr input1, NodePtr input2); 
  void add_gate (int type, NodePtr output, NodePtr input);
  void dump();

  // j-frontier manipulation
  void jf_add (int varID);
  void jf_del (int varID);
  VariablePtr jf_dequeue ();
  int jf_get (int varID);
  void checkout_jf_on_fanouts (int varID);
  void checkin_jf_on_fanouts (int varID);
  void checkout_jf_on_var (int varID);
  void checkin_jf_on_var (int varID);
  void jf_dump();
  int jf_size();
  List <int> &jf() { return jFrontier; }

  // justification counters manipulation 
  void update_fanouts (int varID, int value, int count);

  // netlist handling
  int load_netlist_iscas85 (char *filename);

  // depth calculus
  void calc_depth();

  // clause manipulation
  ClausePtr add_clause() { ClausePtr c = _clDB.add_clause();
                           _reg_clauses.append(c); return c; }
  LiteralPtr add_literal (ClausePtr cl, Node *node, int sign)
    { return _clDB.add_literal(cl, node->variable, sign); }                 

  // I/O manipulation
  void add_PI (NodePtr node) { PIs[PIs.incr_size()] = node;
                               node->type = INPUT; }
  void add_PO (NodePtr node);

  VariablePtr simple_backtrace (int varID, int *value);
  VariablePtr var_imp_clauses (int *value);

  void reset();
  ClauseDatabase &clDB () { return _clDB; }
  
  DArray<ListItem<int> *> *Jf_data() { return &_Jf_data; }

  DNodeArray nodes;
  DNodeArray PIs, POs;

private:

  DArray<ListItem<int> *> _Jf_data;
  List<int> jFrontier;			/* list of ID's of nodes in the JF */
  List<ClausePtr> _reg_clauses;		/* list of regular clauses */
  List<ClausePtr> _imp_clauses;		/* list of implicational clauses */
  ClauseDatabase &_clDB;

};

#endif // __TAU_CIRCUIT__
